from typing import Any
from trism.meta import Metadata
import tritonclient.grpc as grpcclient
import tritonclient.http as httpclient
from tritonclient.grpc import InferenceServerClient as GrpcClient
from tritonclient.http import InferenceServerClient as HttpClient
# NOTE: attrdict broken in python 3.10 and not maintained.
# https://github.com/wallento/wavedrompy/issues/32#issuecomment-1306701776
try:
  from attrdict import AttrDict
except ImportError:
  # Monkey patch collections
  import collections
  import collections.abc
  for type_name in collections.abc.__all__:
    setattr(collections, type_name, getattr(collections.abc, type_name))
  from attrdict import AttrDict


def protoclient(grpc: bool = True) -> Any:
  return grpcclient if grpc else httpclient


def serverclient(
    url: str, 
    grpc: bool = True, 
    concurrency: int = 10, 
    *args, 
    **kwds
) -> GrpcClient | HttpClient:
  return GrpcClient(url=url, *args, **kwds) if grpc else \
    HttpClient(url=url, concurrency=concurrency, *args, **kwds)


def metadata(
  client: GrpcClient | HttpClient,
  model: str,
  version: str = "",
  *args,
  **kwds    
) -> Any:
  meta = client.get_model_metadata(model, version, *args, **kwds)
  conf = client.get_model_config(model, version, *args, **kwds)
  if isinstance(client, GrpcClient):
    conf = conf.config
  else:
    meta, conf = AttrDict(meta), AttrDict(conf)
  inputs = [Metadata(
    name=inp.name,
    shape=inp.shape,
    dtype=inp.datatype
  ) for inp in meta.inputs]
  outputs = [Metadata(
    name=out.name,
    shape=out.shape,
    dtype=out.datatype
  ) for out in meta.outputs]
  return inputs, outputs



from trism import client
import numpy as np


class TritonModel:
  """
  Triton inference server model.
  """

  @property
  def model(self) -> str:
    return self._model
  
  @property
  def version(self) -> str:
    return self._version
  
  @property
  def url(self) -> str:
    return self._url
  
  @property
  def grpc(self) -> bool:
    return self._grpc
  
  @property
  def inputs(self):
    return self._inputs
  
  @property
  def outputs(self):
    return self._outputs

  def __init__(
    self,
    model: str,
    version: int,
    url: str,
    grpc: bool = True
  ) -> None:
    self._url = url
    self._grpc = grpc
    self._model = model
    self._version = str(version) if version > 0 else ""
    self._proto = client.protoclient(grpc)
    self._client = client.serverclient(url, grpc)
    self._inputs, self._outputs = client.metadata(self._client, model, version)
  
  def run(self, data: list[np.array]):
    inputs = [self._inputs[i].make_input(self._proto, data[i]) for i in range(len(self._inputs))]
    outputs = [output.make_output(self._proto) for output in self._outputs]
    results = self._client.infer(self._model, inputs, self._version, outputs)
    return {output.name: results.as_numpy(output.name) for output in self._outputs}



import numpy as np
from trism import types
import tritonclient.grpc as grpcclient
import tritonclient.http as httpclient


class Metadata:
  """
  Keep metadata about inference input/output.
  """

  @property
  def name(self) -> str:
    return self.__name
  
  @property
  def shape(self) -> tuple:
    return self.__shape
  
  @property
  def dtype(self) -> np.dtype:
    return self.__dtype

  def __init__(self, name: str, shape: tuple, dtype: str, *args, **kwds):
    self.__name = name
    self.__shape = shape
    self.__dtype = types.triton_to_np(dtype)

  def make_input(self, client, data: np.array) -> grpcclient.InferInput | httpclient.InferInput:
    infer = client.InferInput(
      name=self.name,
      shape=data.shape,
      datatype=types.np_to_triton(self.dtype)
    )
    infer.set_data_from_numpy(np.astype(data, self.dtype))
    return infer
  
  def make_output(self, client) -> grpcclient.InferRequestedOutput | httpclient.InferRequestedOutput:
    return client.InferRequestedOutput(name=self.name)



import numpy as np

__TRITON_STRING_TO_NUMPY__ = {
  "BOOL": bool,
  "UINT8": np.uint8,
  "UINT16": np.uint16,
  "UINT32": np.uint32,
  "UINT64": np.uint64,
  "INT8": np.int8,
  "INT16": np.int16,
  "INT32": np.int32,
  "INT64": np.int64,
  "FP16": np.float16,
  "FP32": np.float32,
  "FP64": np.float64,
  "STRING": np.object_,
}

__NUMPY_TO_TRITON_STRING__ = {
  bool: "BOOL",
  np.uint8: "UINT8",
  np.uint16: "UINT16",
  np.uint32: "UINT32",
  np.uint64: "UINT64",
  np.int8: "INT8",
  np.int16: "INT16",
  np.int32: "INT32",
  np.int64: "INT64",
  np.float16: "FP16",
  np.float32: "FP32",
  np.float64: "FP64",
  np.object_: "STRING"
}


def triton_to_np(dtype: str) -> np.dtype:
  dtype = dtype.upper()
  if dtype in __TRITON_STRING_TO_NUMPY__:
    return __TRITON_STRING_TO_NUMPY__[dtype]
  elif dtype.startswith("TYPE_"):
    return __TRITON_STRING_TO_NUMPY__[dtype.replace("TYPE_", "")]
  raise TypeError(f"Cannot convert triton type {dtype} to numpy type!")


def np_to_triton(dtype: np.dtype, prefix: bool = False) -> str:
  dtype = __NUMPY_TO_TRITON_STRING__[dtype]
  return f"TYPE_{dtype}" if prefix else dtype



import os
import re
import subprocess
from setuptools import find_packages, setup


def gitversion() -> str:
  _re = re.compile('^Version: (.+)$', re.M)
  d = os.path.dirname(__file__)
  if os.path.isdir(os.path.join(d, '.git')):
    # Get the version using "git describe".
    cmd = 'git describe --tags --match [0-9]*'.split()
    try:
      version = subprocess.check_output(cmd).decode().strip()
    except subprocess.CalledProcessError:
      print('Unable to get version number from git tags')
      exit(1)
    # PEP 386 compatibility
    if '-' in version:
      version = '.post'.join(version.split('-')[:2])
    # Don't declare a version "dirty" merely because a time stamp has changed. 
    # If it is dirty, append a ".dev1" suffix to indicate a development revision after the release.
    with open(os.devnull, 'w') as fd_devnull:
      subprocess.call(['git', 'status'], stdout=fd_devnull, stderr=fd_devnull)
    cmd = 'git diff-index --name-only HEAD'.split()
    try:
      dirty = subprocess.check_output(cmd).decode().strip()
    except subprocess.CalledProcessError:
      print('Unable to get git index status')
      exit(1)
    if dirty != '':
      version += '.dev1'
  else:
    # Extract the version from the PKG-INFO file.
    with open(os.path.join(d, 'PKG-INFO')) as f:
      version = _re.search(f.read()).group(1)
  return version


setup(
  name="trism",
  version=gitversion(),
  packages=find_packages(),
  license="GNU AGPL v3.0",
  zip_safe=True,
  description="Inference with Triton Inference Server easily.",
  long_description=open("README.md", encoding="utf-8").read(),
  long_description_content_type="text/markdown",
  author="Hieu Pham",
  author_email="64821726+hieupth@users.noreply.github.com",
  url="https://github.com/hieupth/trismid",
  install_requires=['pinecore'],
  classifiers=['Intended Audience :: Developers', 'Topic :: Software Development :: Build Tools']
)